/*
 * highlight chat for ii similar in style to irssi.
 *
 * supports:
 * - highlight own nick, words + beep (can set urgent hint).
 * - highlight own messages.
 * - highlight urls.
 * - highlight server messages.
 * - align nicknames.
 * - ignore words.
 * - don't print date-prefix (YYYY-mm-dd).
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

#include <sys/types.h>
#include <regex.h>

/* escape codes: colors, reset */
static const char *green = "\x1b[1;32m";
static const char *yellow = "\x1b[1;33m";
static const char *blue = "\x1b[1;34m";
static const char *white = "\x1b[1;37m";
static const char *reset = "\x1b[1;0m";

/* highlight / color words */
static const char *regexword = "(evil_bob|bob|hiltjo)";
/* match nickname */
static const char *regexnick = "(evil_bob|bob|hiltjo)";
/* nickname to match and align */
static const char *regexnickalign = "^[0-9]{2}:[0-9]{2} <([^>]+)>";
/* match server messages */
static const char *regexserver = "^([0-9]{2}:[0-9]{2} -!- .*)$";
/* skip this prefix for each line (date part) */
static const char *regextimestamp = "^[0-9]{4}-[0-9]{2}-[0-9]{2} ";
/* ignore message on match */
static const char *regexignore = "";
/* highlight / color urls */
static const char *regexurl =
	"(((https?|ftp)|mailto):(//)?[^ <>\"[:blank:]]*|(www|ftp)[0-9]?\\.[-a-z0-9.]+)";

static regex_t nick, nickalign, server, timestamp, ignore, url, word;

static void
eprintf(const char *s)
{
	fprintf(stderr, "%s\n", s);
	exit(EXIT_FAILURE);
}

static void
printalign(const char *s, size_t len, size_t maxlen)
{
	size_t i;

	if(len < maxlen) {
		for(i = 0; i < maxlen - len; i++)
			putchar(' ');
	}
	for(i = 0; i < len && i < maxlen; i++)
		putchar(*s++);
}

int
main(void) {
	/* buffer: should be atleast 512 bytes (IRC RFC) + datetime prefix,
	 * assumes line length is lower than sizeof(buffer). */
	char buffer[1024];
	regmatch_t match, matches[3];
	size_t len;
	const char *p;
	int hasmatch, ret;

	/* compile regex */
	if(regcomp(&nick, regexnick, REG_EXTENDED | REG_ICASE))
		eprintf("invalid regex: nick");
	if(regcomp(&nickalign, regexnickalign, REG_EXTENDED))
		eprintf("invalid regex: nickalign");
	if(regcomp(&server, regexserver, REG_EXTENDED))
		eprintf("invalid regex: server");
	if(regcomp(&timestamp, regextimestamp, REG_EXTENDED))
		eprintf("invalid regex: timestamp");
	if(regcomp(&ignore, regexignore, REG_EXTENDED | REG_ICASE))
		eprintf("invalid regex: ignore");
	if(regcomp(&url, regexurl, REG_EXTENDED | REG_ICASE))
		eprintf("invalid regex: url");
	if(regcomp(&word, regexword, REG_EXTENDED | REG_ICASE))
		eprintf("invalid regex: word");

	while(fgets(buffer, sizeof(buffer), stdin)) {
		p = buffer;

		/* skip date part (YYY-mm-dd) of timestamp */
		if((ret = regexec(&timestamp, p, 1, &match, 0)) != REG_NOMATCH)
			p += match.rm_eo - match.rm_so;

		/* highlight / color server messages */
		if((ret = regexec(&server, p, 1, &match, 0)) != REG_NOMATCH) {
			fputs(green, stdout);
			fputs(p, stdout);
			fputs(reset, stdout);
			fflush(stdout);
			continue;
		}

		/* ignore on match */
		if(regexignore[0] && (ret = regexec(&ignore, p, 1, &match, 0)) != REG_NOMATCH)
			continue;

		/* align nicknames */
		if((ret = regexec(&nickalign, p, 2, matches, 0)) != REG_NOMATCH)
		{
			fwrite(p, 1, matches[1].rm_so, stdout);
			putchar('@');
			p += matches[1].rm_so;
			len = matches[1].rm_eo - matches[1].rm_so;
			/* own nick ?: color nickname part white */
			if((ret = regexec(&nick, p, 1, &match, 0)) != REG_NOMATCH &&
				match.rm_so == 0 && match.rm_eo == len)
			{
				fputs(white, stdout);
				printalign(p, len, 10);
				fputs(reset, stdout);
			} else {
				printalign(p, len, 10);
			}
			p += len;
		}

		hasmatch = 1;
		while(hasmatch) {
			hasmatch = 0;
			/* highlight word */
			if(regexword[0] &&
			   (ret = regexec(&word, p, 1, &match, 0)) != REG_NOMATCH)
			{
				fwrite(p, 1, match.rm_so, stdout);
				fputs(yellow, stdout);
				p += match.rm_so;
				fwrite(p, 1, match.rm_eo - match.rm_so, stdout);
				fputs(reset, stdout);
				p += match.rm_eo - match.rm_so;
				/* beep, can be set in your terminal to set urgent hint */
				putchar('\a');
				hasmatch = 1;
			}
			/* highlight url */
			if((ret = regexec(&url, p, 1, &match, 0)) != REG_NOMATCH) {
				fwrite(p, 1, match.rm_so, stdout);
				fputs(blue, stdout);
				p += match.rm_so;
				fwrite(p, 1, match.rm_eo - match.rm_so, stdout);
				fputs(reset, stdout);
				p += match.rm_eo - match.rm_so;
				hasmatch = 1;
			}
		}
		fputs(p, stdout); /* rest */
		fflush(stdout);
	}

	/* free regex */
	regfree(&word);
	regfree(&nick);
	regfree(&nickalign);
	regfree(&server);
	regfree(&timestamp);
	regfree(&ignore);
	regfree(&url);

	return EXIT_SUCCESS;
}
